"""
All the modules

SPDX-License-Identifier: MIT
"""
from .module import *
# In .std we dynamically load all torch.nn Modules and retrofit them to support loading from
# jsonnet configs. This allows for supporting various versions of Pytorch that may add new modules
# without updating this package and tying it to a specific version of torch. Static analyzers,
# like mypy, do not handle dynamically created classes, without making an explicit plugin
# (which could be a future improvement). Thus we simply import all the base torch.nn Modules
# directly in .std as a hack to get around this issue. That said, since both the .module and
# .std package will define different versions of Module (one that it's configurable, and one
# from the base torch.nn), mypy will complain. Simply ignore that complaint below and make sure
# the .module is imported last so mypy knows uses of Module support methods like from_config.
from .std import *  # type:ignore[misc]



"""
Base classes for the configurable versions of torch.nn modules

SPDX-License-Identifier: MIT
"""
from typing import Callable, ClassVar, Mapping, Sequence
from uuid import UUID, uuid4

from torch import nn

from confactory import Catalog
from confactory.common import TypeList

__all__ = ["Module"]


class Module(nn.Module, Catalog, archetype=nn.Module):
    """
    Base class for all modules, including previously defined modules like those
    in torch.nn modules and newly defined modules.
    """

    SEQUENCE_TYPES: ClassVar[TypeList] = (Sequence, nn.Sequential, nn.ModuleList)
    MAPPING_TYPES: ClassVar[TypeList] = (Mapping, nn.ModuleDict)

    uuid: UUID

    def __init__(self):
        super().__init__()

        # Give every module a unique id. This is useful for data parallel, where we cannot
        # rely on a Module hash, because replicas will have different hashes (but should
        # have the same id)
        self.uuid = uuid4()

    def filtered_modules(self, filter: Callable[[nn.Module], bool] = lambda x: False):
        def gather_modules(module: nn.Module):
            for m in module.children():
                if filter(m):
                    yield m

                yield from gather_modules(m)

        yield from gather_modules(self)

    def filtered_parameters(
        self, filter: Callable[[nn.Module], bool] = lambda x: False
    ):
        for m in self.filtered_modules(filter):
            yield from m.parameters(recurse=True)



"""
Optimizers

SPDX-License-Identifier: MIT
"""
from typing import TYPE_CHECKING

from confactory import Catalog


class Optimizer(Catalog):
    """Base class for configurable optimizers"""


if TYPE_CHECKING:
    from torch.optim import *
else:
    from torch import optim
    from ..nn import retrofit_classes

    retrofitted = retrofit_classes(
        optim,
        optim.Optimizer,
        Optimizer,
    )
    globals().update(retrofitted)

    __all__ = ["Optimizer"] + list(retrofitted.keys())



"""
Initialization routines for torch modules

SPDX-License-Identifier: MIT
"""
import re
from typing import Pattern, Union

import attr
import torch
from torch import nn

from confactory import Catalog, configurable

__all__ = [
    "Initializer",
    "Uniform",
    "Normal",
    "TruncatedNormal",
    "Constant",
    "Ones",
    "Zeros",
    "Eye",
    "Dirac",
    "NonlinearInitializer",
    "XavierUniform",
    "XavierNormal",
    "Orthogonal",
    "KaimingInitializer",
    "KaimingUniform",
    "KaimingNormal",
    "SparseInitializer",
]


@configurable
class Initializer(Catalog):
    """A class that encapsulates an initializer"""

    # Regex of parameters to match
    regex: Pattern = attr.ib(converter=re.compile)

    def match(self, name):
        """Whether or not the name matches the parameter regex"""
        # Use search rather than match since don't usually care about matching
        # the full parameter name, just a substring of the fully qualified parameter
        return self.regex.search(name)

    def __call__(self, parameter: torch.Tensor):
        raise NotImplementedError(f"{type(self)} is an abstract initializer!")


@configurable
class Uniform(Initializer):
    """Uniform initializer"""

    a: float = 0.0
    b: float = 1.0

    def __call__(self, parameter: torch.Tensor):
        nn.init.uniform_(parameter, self.a, self.b)


@configurable
class Normal(Initializer):
    """Normal initializer"""

    mean: float = 0.0
    std: float = 1.0

    def __call__(self, parameter: torch.Tensor):
        nn.init.normal_(parameter, self.mean, self.std)


@configurable
class TruncatedNormal(Initializer):
    """Truncated normal initializer"""

    mean: float = 0.0
    std: float = 1.0
    a: float = -2.0
    b: float = 2.0

    def __call__(self, parameter: torch.Tensor):
        nn.init.trunc_normal_(parameter, self.mean, self.std, self.a, self.b)


@configurable
class Constant(Initializer):
    """Constant initializer"""

    val: float

    def __call__(self, parameter: torch.Tensor):
        nn.init.constant_(parameter, self.val)


@configurable
class Ones(Initializer):
    """Ones initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.ones_(parameter)


@configurable
class Zeros(Initializer):
    """Zeros initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.zeros_(parameter)


@configurable
class Eye(Initializer):
    """Identity matrix initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.eye_(parameter)


@configurable
class Dirac(Initializer):
    """Dirac delta initializer"""

    groups: int = 1

    def __call__(self, parameter: torch.Tensor):
        nn.init.dirac_(parameter, self.groups)


def _calculate_gain(activation_or_float: Union[str, float]) -> float:
    """
    If a string representing an activation function is passed in, compute the
    gain, otherwise return the value passed in.
    """
    if isinstance(activation_or_float, str):
        return nn.init.calculate_gain(activation_or_float)

    return activation_or_float


@configurable
class NonlinearInitializer(Initializer):
    """An initializer that automatically determines initialization based on a nonlinearity"""

    gain: float = attr.ib(default=1.0, converter=_calculate_gain)


@configurable
class XavierUniform(NonlinearInitializer):
    """Xavier uniform initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.xavier_uniform_(parameter, self.gain)


@configurable
class XavierNormal(NonlinearInitializer):
    """Xavier normal initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.xavier_normal_(parameter, self.gain)


@configurable
class Orthogonal(NonlinearInitializer):
    """Orthogonal initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.orthogonal_(parameter, self.gain)


@configurable
class KaimingInitializer(Initializer):
    """Base class for Kaiming initializers"""

    a: float = 0.0
    mode: str = attr.ib(
        default="fan_in", validator=attr.validators.in_(("fan_in", "fan_out"))
    )
    nonlinearity: str = "leaky_relu"


@configurable
class KaimingUniform(KaimingInitializer):
    """Kaiming uniform initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.kaiming_uniform_(parameter, self.a, self.mode, self.nonlinearity)


@configurable
class KaimingNormal(KaimingInitializer):
    """Kaiming normal initializer"""

    def __call__(self, parameter: torch.Tensor):
        nn.init.kaiming_normal_(parameter, self.a, self.mode, self.nonlinearity)


@configurable
class SparseInitializer(Initializer):
    """Sparse initializer"""

    sparsity: float
    std: float = 0.01

    def __call__(self, parameter: torch.Tensor):
        nn.init.sparse_(parameter, self.sparsity, self.std)



"""
Configurable versions of the standard modules in torch.nn

SPDX-License-Identifier: MIT
"""
import inspect
import re
import types
from enum import Enum
# _eval_type is undocumented, but exists starting in python3.6 and is the only way to resolve
# forward references without calling get_type_hints, which does not work when combined with
# the inspect module
from typing import _eval_type  # type: ignore
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Dict, List, Mapping,
                    NamedTuple, Optional, OrderedDict, Sequence, Set, Type,
                    Union, get_type_hints, overload)

import attr
import torch
from torch import nn

from confactory import Catalog, configurable
from confactory.common import Args, Kwargs
from confactory.converters import get_base_type

retrofitted: Dict[str, Type] = {}


class AttributeOverride(NamedTuple):
    """Define an attribute from the super class"""

    name: str
    type: Any = None
    default: Any = attr.NOTHING
    serializer: Optional[Callable] = None


def is_valid_annotation(annotation: Any) -> bool:
    """Checks whether the provided class has an attribute."""
    base_type = get_base_type(annotation)
    return base_type and base_type is not ClassVar


def get_annotation(
    cls: Type,
    name: str,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
) -> Any:
    """Checks whether the provided class has an attribute."""
    annotations = get_type_hints(cls)
    # annotations = getattr(cls, "__annotations__", {})
    if name in annotations:
        return _eval_type(annotations[name], globalns=globalns, localns=localns)


def infer_type(
    param: inspect.Parameter,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
) -> Any:
    """Infer the type of a parameter for a given class"""
    if param.annotation is inspect.Parameter.empty:
        return None

    if param.kind is inspect.Parameter.VAR_POSITIONAL:
        return Args[param.annotation]  # type:ignore

    if param.kind is inspect.Parameter.VAR_KEYWORD:
        return Kwargs[param.annotation]  # type:ignore

    return _eval_type(param.annotation, globalns=globalns, localns=localns)


def get_default(cls, name) -> Any:
    """Checks the default value of an __init__ parameter."""
    signature = inspect.signature(cls)
    if name in signature.parameters:
        parameter = signature.parameters[name]

        if parameter.default is inspect.Parameter.empty:
            return attr.NOTHING

        return parameter.default

    return getattr(cls, name, attr.NOTHING)


if TYPE_CHECKING:
    from torch.nn.modules import *  # noqa: F401,F403

    def retrofit(
        maybe_cls=None,
        *,
        overrides: Sequence[AttributeOverride] = tuple(),
        repr: bool = False,
    ):
        """Automatically make a torch class confgurable"""

    def retrofit_classes(
        module,
        to_retrofit: Type,
        catalog_cls: Type[Catalog],
        exclusions: Optional[Set[str]] = None,
        overrides: Optional[Mapping[str, Sequence[AttributeOverride]]] = None,
        repr: bool = False,
    ) -> Dict[str, Type]:
        """
        Retrofit all subclasses of to_retrofit as catalog_cls from the passed in module and
        return a dict of names to retrofitted classes
        """

else:
    from .module import Module

    def retrofit(
        maybe_cls=None,
        *,
        overrides: Sequence[AttributeOverride] = tuple(),
        repr: bool = False,
    ):
        """Automatically make a torch module confgurable"""

        def wrap(cls):
            signature = inspect.signature(cls)
            these = {
                name: attr.ib(
                    repr=repr,
                    init=False,
                    metadata={
                        "super_init": True,
                        "super_default": get_default(cls, name),
                    },
                    type=infer_type(param, globalns=vars(nn))
                    or get_annotation(cls, name, globalns=vars(nn)),
                )
                for name, param in signature.parameters.items()
            }

            if repr and hasattr(cls, "__annotations__"):
                # Add any remaining attributes that get set automatically by
                # the super cls __init__
                these.update(
                    {
                        name: attr.ib(
                            repr=repr,
                            init=False,
                            metadata={
                                "super_init": False,
                                "super_default": get_default(cls, name),
                            },
                            type=type_,
                        )
                        for name, type_ in getattr(cls, "__annotations__").items()
                        if name not in these and is_valid_annotation(type_)
                    }
                )

            # TODO: Add a serializer to AttributeOverride to ensure nn.Sequential works
            for override in overrides:
                try:
                    attribute = these[override.name]
                except KeyError as exc:
                    raise ValueError(f"Cannot override {override.name}") from exc

                if override.type:
                    attribute.type = override.type

                if override.default is not attr.NOTHING:
                    attribute.metadata["super_default"] = override.default

                if override.serializer is not None:
                    attribute.metadata["serializer"] = override.serializer

            return configurable(
                cls,
                repr=repr,
                these=these,  # only the annotations on the class
                on_setattr=attr.setters.NO_OP,  # no additional setters
            )

        if maybe_cls:
            return wrap(maybe_cls)

        return wrap

    def retrofit_classes(
        module,
        to_retrofit: Type,
        catalog_cls: Type[Catalog],
        exclusions: Optional[Set[str]] = None,
        overrides: Optional[Mapping[str, Sequence[AttributeOverride]]] = None,
        repr: bool = False,
    ) -> Dict[str, Type]:
        """
        Retrofit all subclasses of to_retrofit as catalog_cls from the passed in module and
        return a dict of names to retrofitted classes
        """
        overrides = overrides or {}
        exclusions = exclusions or set()

        retrofitted: Dict[str, Type] = {}
        for name, sub_cls in vars(module).items():
            if not isinstance(sub_cls, type(to_retrofit)):
                continue

            if name in exclusions:
                continue

            if sub_cls is to_retrofit or not issubclass(sub_cls, to_retrofit):
                continue

            retrofitted[name] = retrofit(
                types.new_class(name, (sub_cls, catalog_cls), {"archetype": sub_cls}),
                overrides=overrides.get(name, tuple()),
                repr=repr
            )

        return retrofitted


    BiasOverride = AttributeOverride(
        "bias",
        serializer=lambda self, *args: self.bias is not None
    )

    RETROFIT_EXCLUSIONS = {"Container"}
    RETROFIT_OVERRIDES = {
        "Sequential": [
            AttributeOverride(
                "args",
                type=Union[Args[Module], OrderedDict[str, Module]],
                serializer=lambda self, *args: [m for m in self]
            )
        ],
        "Linear": [BiasOverride],
        "Bilinear": [BiasOverride],
        "LazyLinear": [BiasOverride],
    }

    retrofitted = retrofit_classes(
        nn,
        nn.Module,
        Module,
        exclusions=RETROFIT_EXCLUSIONS,
        overrides=RETROFIT_OVERRIDES,
    )
    globals().update(retrofitted)
    Sequential = retrofitted["Sequential"]


@configurable
class Parallel(Sequential):
    """Run a list of modules in parallel and return their outputs"""

    def forward(self, input):
        return [module(input) for module in self]

    def __repr__(self):
        return super().__repr__()


@configurable
class Select(Module):
    field: str

    def forward(self, input):
        return getattr(input, self.field)


@configurable
class Apply(Module):
    fields: str
    module: Module

    def forward(self, input):
        for field in self.fields.split(","):
            attr = None
            for subfield in field.split("."):
                index = None
                parent = attr or input
                match = re.match(r"(\w+)(?:\[([0-9]+)\])", subfield)
                if match:
                    subfield, index = match.group(1, 2)

                attr = getattr(parent, subfield)
                if index is not None:
                    parent = attr
                    index = int(index)
                    attr = attr[index]

            if index is not None:
                parent[index] = self.module(attr)
            else:
                setattr(parent, subfield, self.module(attr))

        return input


class ReduceOp(str, Enum):
    """Types of reduction operations"""

    MIN = "MIN"
    MAX = "MAX"
    SUM = "SUM"
    PROD = "PROD"


@configurable
class Reduce(Module):
    """Perform a reduction on the list of tensors passed in"""

    op: ReduceOp

    @overload
    def forward(self, *args: torch.Tensor) -> None:
        ...

    @overload
    def forward(self, arg: List[torch.Tensor]) -> None:
        ...

    def forward(self, *args):
        stacked = (
            torch.stack(args[0])
            if len(args) == 1 and isinstance(args[0], (list, tuple))
            else torch.stack(args)
        )

        if self.op == ReduceOp.MIN:
            return torch.min(stacked, dim=0)

        if self.op == ReduceOp.MAX:
            return torch.max(stacked, dim=0)

        if self.op == ReduceOp.SUM:
            return torch.sum(stacked, dim=0)

        if self.op == ReduceOp.PROD:
            return torch.prod(stacked, dim=0)

        raise ValueError(f"Unknown reduction op {self.op}!")


__all__ = [
    "retrofit",
    "retrofit_classes",
    "AttributeOverride",
    "Parallel",
    "Select",
    "Apply",
    "Reduce",
    "ReduceOp",
] + list(retrofitted.keys())



"""
Learning rate schedules

SPDX-License-Identifier: MIT
"""
from typing import TYPE_CHECKING, Callable, Sequence, Union

from confactory import Catalog, configurable


retrofitted: Dict[str, Type] = {}


class LRScheduler(lr_scheduler.LRScheduler, Catalog, archetype=lr_scheduler.LRScheduler):
    """Base class for configurable learning rate schedulers"""


class LRLambda(Catalog):
    """Base class for configurable learning rate lambda functions"""


LRCallable = Union[LRLambda, Callable[[int], float]]

if TYPE_CHECKING:
    from torch.optim.lr_scheduler import *  # noqa: F401,F403

else:
    from torch.optim import lr_scheduler
    from ..nn import AttributeOverride, retrofit_classes

    RETROFIT_OVERRIDES = {
        "LambdaLR": [
            AttributeOverride("lr_lambda", Union[LRCallable, Sequence[LRCallable]])
        ]
    }

    retrofitted = retrofit_classes(
        lr_scheduler,
        lr_scheduler._LRScheduler,
        LRScheduler,
        overrides=RETROFIT_OVERRIDES,
        repr=True
    )
    globals().update(retrofitted)


@configurable
class Linear(LRLambda):
    """
    Implements a linear learning rate schedule. Since learning rate schedulers in Pytorch want a
    mutiplicative factor we have to use a non-intuitive computation for linear annealing.

    This needs to be a top-level class in order to pickle it.
    """

    steps: int
    initial_lr: float
    final_lr: float

    def __attrs_post_init__(self):
        self.lr_step = (self.final_lr - self.initial_lr) / self.steps

    def __call__(self, step: int):
        """The actual learning rate schedule"""
        # Compute the what the previous learning rate should be
        prev_lr = self.initial_lr - step * self.lr_step

        # Calculate the current multiplicative factor
        return (prev_lr + (step + 1) * self.lr_step) / prev_lr


@configurable
class Warmup(LRLambda):
    """
    Implement the learning rate schedule from Attention is All You Need

    This needs to be a top-level class in order to pickle it.
    """

    warmup_steps: int = 0

    def __call__(self, step: int):
        """The actual learning rate schedule"""
        # the schedule doesn't allow for step to be zero (it's raised to the negative power),
        # but the input step is zero-based so just do a max with 1
        step = max(1, step)
        if step < self.warmup_steps:
            return step * self.warmup_steps ** -1.5

        return step ** -0.5


__all__ = [
    "LRScheduler",
    "LRLambda",
    "Linear",
    "Warmup",
] + list(retrofitted.keys())
